"""Handler for exceptions thrown by pubsub listeners."""
import logging
from . import pub

_LOGGER = logging.getLogger(__name__)


class ListenerExceptionHandler(pub.IListenerExcHandler):
    """Handle exceptions from listeners of pubsub topics."""

    def __call__(self, listenerID, topicObj):
        """Call the exception handler."""
        _LOGGER.error(
            "pubsub listener Exception: %s with topic %s", listenerID, topicObj.name
        )



"""Utility methods."""
from typing import Iterable

from .address import Address
from .constants import (
    HC_LOOKUP,
    UC_LOOKUP,
    X10Commands,
    X10CommandType,
    ResponseStatus,
    MessageFlagType,
)
from .protocol.commands import commands


def housecode_to_byte(housecode: str) -> int:
    """Return the byte value of an X10 housecode."""
    return HC_LOOKUP.get(housecode.lower())


def unitcode_to_byte(unitcode: int) -> int:
    """Return the byte value of an X10 unitcode."""
    return UC_LOOKUP.get(unitcode)


def byte_to_housecode(bytecode: int) -> str:
    """Return an X10 housecode value from a byte value."""
    house_code = list(HC_LOOKUP.keys())[list(HC_LOOKUP.values()).index(bytecode)]
    return house_code.upper()


def byte_to_unitcode(bytecode: int) -> int:
    """Return an X10 unitcode value from a byte value."""
    return list(UC_LOOKUP.keys())[list(UC_LOOKUP.values()).index(bytecode)]


def byte_to_int(bytecode: bytes) -> int:
    """Return an int from a byte string."""
    return int.from_bytes(bytecode, byteorder="big")


def x10_command_type(command: X10Commands) -> X10CommandType:
    """Return the X10 command type from an X10 command."""
    command_type = X10CommandType.DIRECT
    cmd_val = command.value if isinstance(command, X10Commands) else command
    if cmd_val in [
        X10Commands.ALL_UNITS_OFF.value,
        X10Commands.ALL_LIGHTS_ON.value,
        X10Commands.ALL_LIGHTS_OFF.value,
    ]:
        command_type = X10CommandType.BROADCAST
    return command_type


def raw_x10_to_bytes(raw_x10: int) -> int:
    """Return the byte value of a raw X10 command."""
    yield raw_x10 >> 4
    yield raw_x10 & 0x0F


def bit_is_set(bitmask: int, bit: int) -> bool:
    """Return True if a specific bit is set in a bitmask.

    Uses the low bit is 1 and the high bit is 8.
    """
    return bool(bitmask & (1 << bit))


def set_bit(data: int, bit: int, is_on: bool) -> int:
    """Set the value of a bit in a bitmask on or off.

    Uses the low bit is 0 and the high bit is 7.
    """
    if isinstance(data, bytes):
        bitmask = int.from_bytes(data, byteorder="big")
    else:
        bitmask = data
    if is_on:
        return bitmask | (1 << bit)
    return bitmask & (0xFF & ~(1 << bit))


def vars_to_bytes(vals: Iterable) -> bytes:
    """Create a byte string from a set of values."""
    msg = bytearray()
    for val in vals:
        if val is None:
            pass
        elif isinstance(val, (int, bytes)):
            msg.extend(bytes([val]))
        else:
            msg.extend(bytes(val))
    return bytes(msg)


def vars_to_string(vals: Iterable) -> str:
    """Create a byte string from a set of values."""
    from enum import Enum, IntEnum

    output = []
    for fld, val in vals:
        if val is None:
            pass
        elif isinstance(val, (Enum, IntEnum)):
            valstr = str(val)
        elif isinstance(val, (int, bytes)):
            valstr = "0x{0:02x}".format(val)
        else:
            valstr = str(val)
        output.append("{}: {}".format(fld, valstr))
    return ", ".join(output)


def vars_to_repr(vals: Iterable) -> str:
    """Create a byte string from a set of values."""
    from enum import Enum, IntEnum

    output = []
    for fld, val in vals:
        if val is None:
            pass
        elif isinstance(val, (Enum, IntEnum)):
            valstr = repr(val)
        elif isinstance(val, (int, bytes)):
            valstr = "0x{0:02x}".format(val)
        else:
            valstr = repr(val)
        output.append("{}: {}".format(fld, valstr))
    return ", ".join(output)


def test_values_eq(val1, val2) -> bool:
    """Test if val1 eq val2 for template management."""
    if val1 is None or val2 is None:
        return True
    if val1 == val2:
        return True
    return False


def _include_address(prefix, topic, address, message_type):
    """Test if we should include the address in the topic build."""
    if address is None:
        return False

    if isinstance(message_type, str):
        message_type = getattr(MessageFlagType, str(message_type).upper())

    if prefix == "send" and message_type == MessageFlagType.DIRECT:
        return False

    # if message_type in [MessageFlagType.ALL_LINK_CLEANUP,
    #                     MessageFlagType.DIRECT_ACK,
    #                     MessageFlagType.DIRECT_NAK]:
    #     return False

    if commands.get(topic) is None:
        return False

    return True


def _include_group(topic, group, message_type):
    if group is None:
        return False

    if not commands.use_group(topic):
        return False

    if message_type in [MessageFlagType.ALL_LINK_CLEANUP]:
        return False

    return True


def build_topic(topic, prefix=None, address=None, group=None, message_type=None):
    """Build a full topic from components."""
    full_topic = ""
    if prefix is not None:
        # Adding the . separator since there must be something after a prefix
        full_topic = "{}.".format(str(prefix))

    if _include_address(prefix, topic, address, message_type):
        addr = address.id if isinstance(address, Address) else address
        full_topic = "{}{}.".format(full_topic, addr)
        if commands.use_group(topic) and group is not None:
            group = group if group else 1
            full_topic = "{}{}.".format(full_topic, group)

    full_topic = "{}{}".format(full_topic, topic)
    if message_type is not None:
        full_topic = "{}.{}".format(full_topic, str(message_type))
    return full_topic


def multiple_status(*args):
    """Return the proper status based on the worst case of all status responses."""
    worst_response = 1
    for response in args:
        if response is None:
            continue
        if int(response) == 0 or worst_response == 0:
            worst_response = 0
        elif int(response) > worst_response:
            worst_response = int(response)
    return ResponseStatus(worst_response)



"""Protocol classes to interface with serial, socket and http devices."""
import asyncio
import logging
from functools import partial

import serial

from .serial_transport import SerialTransport

_LOGGER = logging.getLogger(__name__)


def topic_to_message_type(topic):
    """Return MessageFlagType from the topic."""
    from ..constants import MessageFlagType

    subtopics = topic.name.split(".")
    flag = "direct" if len(subtopics) < 3 else subtopics[2]
    for flag_type in MessageFlagType:
        if flag.lower() == str(flag_type):
            return flag_type
    return MessageFlagType(0)


def topic_to_message_handler(topic, register_list):
    """Register handler to topic."""

    def register(func):
        register_list["send.{}".format(topic)] = func
        return func

    return register


def topic_to_command_handler(topic, register_list):
    """Register handler to topic."""

    def register(func):
        register_list["send.{}".format(topic)] = func
        return func

    return register


async def async_connect_serial(device, protocol):
    """Connect to the PowerLine Modem via serial port.

    Parameters:
        port – Device name.
        protocol - Insteon Modem Protocol instance.

    """
    loop = asyncio.get_event_loop()
    try:
        ser = serial.serial_for_url(url=device, baudrate=19200)
        transport = SerialTransport(loop, protocol, ser, device=device)
    except OSError as ex:
        _LOGGER.warning("Unable to connect to %s: %s", device, ex)
        transport = None
    return transport


async def async_connect_socket(host, protocol, port=None):
    """Connect to the Hub Version 1 via TCP Socket."""

    port = 9761 if not port else port
    loop = asyncio.get_event_loop()
    url = "socket://{}:{}".format(host, port)
    ser = serial.serial_for_url(url=url, baudrate=19200)
    transport = SerialTransport(loop, protocol, ser, device=url)
    return transport


async def async_connect_http(host, username, password, protocol, port=None):
    """Connect to the Hub Version 2 via HTTP."""
    from .http_transport import HttpTransport

    port = 25105 if not port else port
    transport = HttpTransport(
        protocol=protocol, host=host, port=port, username=username, password=password
    )
    if await transport.async_test_connection():
        protocol.connection_made(transport)
    return transport


async def async_modem_connect(
    device=None, host=None, port=None, username=None, password=None, hub_version=2
):
    """Connect to the Insteon Modem.

    Returns an Insteon Modem object (PLM, Hub, or Hub1)

    Parameters:
        device: Serial port device (i.e. /dev/ttyUSB0 or COM5)
        host: Hub IP address (i.e. 192.168.1.100)
        port: Hub port number (Default 25105 for version 2 or 9761 for version 1)
        username: Hub username for the Hub V2
        password: Hub password for the Hub V2
        hub_version: 1 | 2 (Default: 2)

    If the device is a serial device see the serial class parameters.

    """
    from .protocol import Protocol
    from ..handlers.get_im_info import GetImInfoHandler

    modem_address = "000000"
    modem_cat = 0x03
    modem_subcat = 0x00
    modem_firmware = 0x00

    def set_im_info(address, cat, subcat, firmware):
        nonlocal modem_address, modem_cat, modem_subcat, modem_firmware
        modem_address = address
        modem_cat = cat
        modem_subcat = subcat
        modem_firmware = firmware

    transport = None
    if not device and not host:
        return ValueError("Must specify either a device or a host")

    if device:
        from ..device_types.plm import PLM as Modem

        connect_method = partial(async_connect_serial, **{"device": device})
        protocol = Protocol(connect_method=connect_method)

    elif hub_version == 2:
        from ..device_types.hub import Hub as Modem

        connect_method = partial(
            async_connect_http,
            **{"host": host, "username": username, "password": password, "port": port}
        )
        protocol = Protocol(connect_method=connect_method)

    else:
        from ..device_types.plm import PLM as Modem

        connect_method = partial(
            async_connect_socket, **{"host": host, "port": port, "protocol": protocol}
        )
        protocol = Protocol(connect_method=connect_method)

    await protocol.async_connect()

    get_im_info = GetImInfoHandler()
    get_im_info.subscribe(set_im_info)
    # TODO check for success or failure
    await get_im_info.async_send()
    modem = Modem(
        address=modem_address,
        cat=modem_cat,
        subcat=modem_subcat,
        firmware=modem_firmware,
    )
    modem.protocol = protocol
    modem.transport = transport
    # Pause to allow connection_made to be called:
    return modem



"""Wrapper for serial.aio transport."""
from serial_asyncio import SerialTransport as SerialTransportBase
from serial import SerialException


class SerialTransport(SerialTransportBase):
    """Wrapper for serial_asyncio.SerialTransport."""

    def __init__(self, loop, protocol, serial_instance, device):
        """Init the SerialTransport class."""
        super().__init__(loop=loop, protocol=protocol, serial_instance=serial_instance)
        self._device = device

    @property
    def connected(self):
        """Return True if the transport is connected to the serial device."""
        return not self._closing

    def write(self, data):
        """Override SerialTransport write method."""
        msg_bytes = bytes(data)
        try:
            super().write(msg_bytes)
        except SerialException as exc:
            self._protocol.connection_lost(exc)

    async def async_write(self, data):
        """Asyncronous write method."""
        self.write(data)

    async def async_connect(self, protocol, device=None, loop=None):
        """Connect to a serial device asycrounously."""
        from . import async_connect_serial

        device = device if device else self._device
        transport = async_connect_serial(device=device, protocol=protocol)
        return transport

    def _poll_read(self):
        try:
            super()._poll_read()
        except SerialException as exc:
            self._protocol.connection_lost(exc)



"""General controller devices (cat: 0x00)."""
from .battery_base import BatteryDeviceBase
from .device_base import Device
from .variable_controller_base import VariableControllerBase


class GeneralController(Device):
    """General controller device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController class."""
        from ..aldb.no_aldb import NoALDB

        super().__init__(address, cat, subcat, firmware, description, model)
        self._aldb = NoALDB(self._address)


class GeneralController_ControlLinc(VariableControllerBase):
    """ControLinc 2430 device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_ControlLinc class."""
        buttons = {1: None, 2: None, 3: None, 4: None, 5: None, 6: None}
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )


class GeneralController_RemoteLinc(BatteryDeviceBase, VariableControllerBase):
    """RemoteLinc 2440 device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_RemoteLinc class."""
        buttons = {1: None, 2: None, 3: None, 4: None, 5: None, 6: None}
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )

    def _register_operating_flags(self):
        from ..operating_flag import STAY_AWAKE_ON

        super()._register_operating_flags()
        self._add_operating_flag(STAY_AWAKE_ON, 0, 3, 6, 7)


class GeneralController_MiniRemoteBase(BatteryDeviceBase, VariableControllerBase):
    """RemoteLinc 2440 device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the GeneralController_MiniRemoteBase class."""
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )
        self._database_delta = 0

    def _register_operating_flags(self):
        from ..operating_flag import (
            PROGRAM_LOCK_ON,
            STAY_AWAKE_ON,
            SEND_ON_ONLY,
            KEY_BEEP_ON,
            LED_ON,
            GROUPED_ON,
        )

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_ON, 0, 1, 2, 3)
        self._add_operating_flag(KEY_BEEP_ON, 0, 2, 4, 5)
        self._add_operating_flag(STAY_AWAKE_ON, 0, 3, 6, 7)
        self._add_operating_flag(SEND_ON_ONLY, 0, 4, 8, 9)
        self._add_operating_flag(GROUPED_ON, 0, 6, 0x0E, 0x0F)


# TODO setup ability to swich from toggle to button groups
class GeneralController_MiniRemote_Switch(GeneralController_MiniRemoteBase):
    """RemoteLinc 2440 device with a single button."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_MiniRemote_Switch class."""
        buttons = {1: None, 2: None}
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )


class GeneralController_MiniRemote_4(GeneralController_MiniRemoteBase):
    """RemoteLinc 2440 device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_MiniRemote_4 class."""
        buttons = {1: None, 2: None, 3: None, 4: None}
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )


class GeneralController_MiniRemote_8(GeneralController_MiniRemoteBase):
    """RemoteLinc 2440 device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_MiniRemote_8 class."""
        buttons = {
            1: None,
            2: None,
            3: None,
            4: None,
            5: None,
            6: None,
            7: None,
            8: None,
        }
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )



"""Send a Start All-Linking command."""
from .outbound_base import OutboundHandlerBase
from ..topics import SEND_ALL_LINK_COMMAND
from ..constants import AllLinkMode
from . import ack_handler


class SendAllLinkingCommandHandler(OutboundHandlerBase):
    """Handle Start All-LInking commands."""

    def __init__(self):
        """Init the StartAllLinking class."""
        super().__init__(SEND_ALL_LINK_COMMAND)

    # pylint: disable=arguments-differ
    def send(self, group: int = 0, mode: AllLinkMode = AllLinkMode.CONTROLLER):
        """Send the Start All-Linking message."""
        super().send(group=group, mode=mode)

    # pylint: disable=arguments-differ
    async def async_send(
        self, group: int = 0, mode: AllLinkMode = AllLinkMode.CONTROLLER
    ) -> bool:
        """Send the Start All-Linking message asyncronously."""
        return await super().async_send(group=group, mode=mode)

    @ack_handler(wait_response=False)
    def handle_ack(self, group, mode):
        """Handle the message ACK returning True to the async_send method."""



"""Manage X10 functions."""


def all_units_off():
    """Send the X10 All Units OFF command."""


def all_lights_on():
    """Send the X10 All Lights ON command."""


def all_lights_off():
    """Send the X10 All Lights OFF command."""



"""Mock message handling via Chain of Command pattern."""
import asyncio
from abc import ABCMeta
from .. import pub
from .inbound_base import InboundHandlerBase
from . import ResponseStatus
from ..utils import build_topic


# It should not take more than 3 minutes for a message send (I hope)
TIMEOUT = 60 * 3


class OutboundHandlerBase(InboundHandlerBase):
    """Manage a message chain."""

    __meta__ = ABCMeta

    def __init__(self, topic, address=None, group=None, message_type=None):
        """Init the MessageManager."""
        self._message_response = asyncio.Queue()
        self._message_type = message_type
        super().__init__(topic, address=address, group=group, message_type=message_type)

    @property
    def message_response(self) -> asyncio.Queue:
        """Message response queue to manage message status."""
        return self._message_response

    def send(self, **kwargs):
        """Sent the message."""
        asyncio.ensure_future(self.async_send(**kwargs))

    async def async_send(self, **kwargs):
        """Send the message and wait for a status."""
        while not self._message_response.empty():
            # Empty the message queue
            try:
                self._message_response.get_nowait()
            except asyncio.QueueEmpty:
                pass
        send_topic = build_topic(
            topic=self._topic,
            prefix="send",
            address=None,
            group=None,
            message_type=self._message_type,
        )
        pub.sendMessage(send_topic, **kwargs)
        try:
            test = await asyncio.wait_for(self._message_response.get(), TIMEOUT)
            return test
        except asyncio.TimeoutError:
            return ResponseStatus.UNSENT



"""Base class for all subscribable classes."""
from abc import ABC
from typing import Callable
from . import pub


class SubscriberBase(ABC):
    """Event class to manage triggering of events."""

    def __init__(self, subscriber_topic):
        """Init the Event class."""
        self._subscriber_topic = subscriber_topic

    def subscribe(self, callback: Callable):
        """Subscribe to the event."""
        topic_mgr = pub.getDefaultTopicMgr()
        topic = topic_mgr.getOrCreateTopic(self._subscriber_topic)
        if not pub.isSubscribed(callback, topicName=topic.name):
            pub.subscribe(callback, topicName=topic.name)

    def unsubscribe(self, callback):
        """Unsubscribe to the event."""
        pub.unsubscribe(callback, topicName=self._subscriber_topic)

    def _call_subscribers(self, **kwargs):
        """Call subscribers to the event."""
        pub.sendMessage(topicName=self._subscriber_topic, **kwargs)



"""Manage saving and restoring devices from JSON file."""
import json
import logging

from ..address import Address
from .device_manager import create_device, DeviceId

DEVICE_INFO_FILE = "insteon_devices.json"
_LOGGER = logging.getLogger(__name__)


def _dict_to_device(device_dict):
    address = Address(device_dict.get("address"))
    aldb_status = device_dict.get("aldb_status", 0)
    aldb = device_dict.get("aldb", {})
    cat = device_dict.get("cat")
    subcat = device_dict.get("subcat")
    firmware = device_dict.get("firmware")
    operating_flags = device_dict.get("operating_flags", {})
    properties = device_dict.get("properties", {})
    device_id = DeviceId(address, cat, subcat, firmware)
    device = create_device(device_id)
    if device:
        aldb_records = _dict_to_aldb_record(aldb)
        device.aldb.load_saved_records(aldb_status, aldb_records)
        for flag in operating_flags:
            value = operating_flags[flag]
            if device.operating_flags.get(flag):
                device.operating_flags[flag].load(value)
        for flag in properties:
            value = properties[flag]
            if device.properties.get(flag):
                device.properties[flag].load(value)
    return device


def _device_to_dict(device_list):
    """Convert a device to a dictionary."""
    from ..x10_address import X10Address

    devices = []
    for addr in device_list:
        device = device_list.get(addr)
        if not isinstance(device.address, X10Address):
            aldb = {}
            for mem in device.aldb:
                rec = device.aldb[mem]
                if rec:
                    aldb_rec = {
                        "memory": mem,
                        "in_use": rec.is_in_use,
                        "controller": rec.is_controller,
                        "high_water_mark": rec.is_high_water_mark,
                        "bit5": rec.is_bit5_set,
                        "bit4": rec.is_bit4_set,
                        "group": rec.group,
                        "target": rec.target.id,
                        "data1": rec.data1,
                        "data2": rec.data2,
                        "data3": rec.data3,
                    }
                    aldb[mem] = aldb_rec
            operating_flags = {}
            for flag in device.operating_flags:
                operating_flags[flag] = device.operating_flags[flag].value
            properties = {}
            for flag in device.properties:
                properties[flag] = device.properties[flag].value
            device_info = {
                "address": device.address.id,
                "cat": device.cat,
                "subcat": device.subcat,
                "firmware": device.firmware,
                "aldb_status": device.aldb.status.value,
                "aldb": aldb,
                "operating_flags": operating_flags,
                "properties": properties,
            }
            devices.append(device_info)
    return devices


def _dict_to_aldb_record(aldb_dict):
    from ..aldb.aldb_record import ALDBRecord

    records = {}
    for mem_addr in aldb_dict:
        rec = aldb_dict[mem_addr]
        memory = int(mem_addr)
        control_flags = int(rec.get("control_flags", 0))
        in_use = rec.get("in_use", bool(control_flags & 1 << 7))
        controller = rec.get("controller", bool(control_flags & 1 << 6))
        bit5 = rec.get("bit5", bool(control_flags & 1 << 5))
        bit4 = rec.get("bit4", bool(control_flags & 1 << 4))
        high_water_mark = rec.get("high_water_mark", not bool(control_flags & 1 << 1))
        group = int(rec.get("group", 0))
        target = rec.get("target", rec.get("address", "000000"))
        data1 = int(rec.get("data1", 0))
        data2 = int(rec.get("data2", 0))
        data3 = int(rec.get("data3", 0))
        records[memory] = ALDBRecord(
            memory=memory,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
    return records


class SavedDeviceManager:
    """Manage saving and restoring devices from JSON."""

    def __init__(self, workdir):
        """Init the SavedDeviceManager class."""
        self._workdir = workdir

    async def async_save(self, device_list: dict):
        """Save all devices to the `insteon_devices.json` file for faster loading."""
        device_dict = _device_to_dict(device_list)
        await self._write_saved_devices(device_dict)

    async def async_load(self) -> {}:
        """Load devices from the saved device file."""
        from .. import devices

        modem = devices.modem
        saved_devices = await self._read_saved_devices()
        devices = {}
        for saved_device in saved_devices:
            address = Address(saved_device.get("address"))
            if address != modem.address:
                device = _dict_to_device(saved_device)
                if device:
                    devices[address] = device
                    _LOGGER.debug(
                        "Device with id %s added to device list "
                        "from saved device data.",
                        address,
                    )
            else:
                aldb_status = saved_device.get("aldb_status", 0)
                aldb = saved_device.get("aldb", {})
                aldb_records = _dict_to_aldb_record(aldb)
                modem.aldb.load_saved_records(aldb_status, aldb_records)
        return devices

    async def _read_saved_devices(self):
        """Load device information from the device info file."""
        _LOGGER.debug("Loading saved device info.")
        from aiofile import AIOFile
        from os import path

        saved_devices = []
        if not self._workdir:
            _LOGGER.debug("Really Loading saved device info.")
            return saved_devices

        try:
            device_file = path.join(self._workdir, DEVICE_INFO_FILE)
            async with AIOFile(device_file, "r") as afp:
                json_file = ""
                json_file = await afp.read()
            try:
                saved_devices = json.loads(json_file)
            except json.decoder.JSONDecodeError:
                _LOGGER.debug("Loading saved device file failed")
        except FileNotFoundError:
            _LOGGER.debug("Saved device file not found")
        return saved_devices

    async def _write_saved_devices(self, devices):
        from aiofile import AIOFile
        from os import path

        _LOGGER.debug("Writing %d devices to save file", len(devices))
        device_file = path.join(self._workdir, DEVICE_INFO_FILE)
        try:
            async with AIOFile(device_file, "w") as afp:
                out_json = json.dumps(devices, indent=2)
                await afp.write(out_json)
                await afp.fsync()
        except FileNotFoundError as ex:
            _LOGGER.error("Cannot write to file %s", device_file)
            _LOGGER.error("Exception: %s", str(ex))



"""Send an All-Link Cleanup ACK message to a device."""

import asyncio
from ..outbound_base import OutboundHandlerBase
from ...address import Address
from .. import ack_handler
from ...constants import MessageFlagType


class AllLinkCleanupAckCommandHandlerBase(OutboundHandlerBase):
    """Abstract base class for outbound All-Link Cleanup ACK message handling."""

    def __init__(self, topic, address, group):
        """Init the DirectCommandHandlerBase class."""
        self._address = Address(address)
        self._group = group
        self._response_lock = asyncio.Lock()
        super().__init__(
            topic=topic,
            address=self._address,
            group=self._group,
            message_type=MessageFlagType.ALL_LINK_CLEANUP_ACK,
        )

    async def async_send(self, **kwargs):
        """Send the command and wait for a direct_nak."""
        return await super().async_send(address=self._address, **kwargs)

    # pylint: disable=arguments-differ
    @ack_handler(wait_response=False)
    def handle_ack(self, cmd1, cmd2, user_data):
        """Handle the message ACK."""



"""All-Link Database for devices with no All-Link Database."""
from typing import Callable

from ..address import Address
from . import ALDBBase
from .aldb_status import ALDBStatus


class NoALDB(ALDBBase):
    """All-Link Database for devices with no All-Link Database."""

    def __setitem__(self, mem_addr, record):
        """Do nothing."""

    @property
    def is_loaded(self) -> bool:
        """Return loaded always."""
        return ALDBStatus.LOADED

    # pylint: disable=arguments-differ
    async def async_load(
        self,
        mem_addr: int = 0x00,
        num_recs: int = 0x00,
        refresh: bool = False,
        callback: Callable = None,
    ):
        """Load the All-Link Database."""
        if callback:
            callback()
        return ALDBStatus.LOADED

    async def async_write(self):
        """Write modified records to the device."""
        return 0

    def add(
        self,
        group: int,
        target: Address,
        controller: bool = False,
        data1: int = 0x00,
        data2: int = 0x00,
        data3: int = 0x00,
    ):
        """Add an All-Link record."""

    def remove(self, mem_addr: int):
        """Remove an All-Link record."""



"""Insteon All-Link Database.

The All-Link database contains database records that represent links to other
Insteon devices that either respond to or control the current device.
"""
from abc import ABC, abstractmethod
import asyncio
import logging
from typing import Callable

from ..address import Address
from .aldb_status import ALDBStatus
from .aldb_version import ALDBVersion
from .aldb_record import ALDBRecord
from ..topics import (
    DEVICE_LINK_CONTROLLER_CREATED,
    DEVICE_LINK_RESPONDER_CREATED,
    DEVICE_LINK_CONTROLLER_REMOVED,
    DEVICE_LINK_RESPONDER_REMOVED,
)
from .. import pub


_LOGGER = logging.getLogger(__name__)


class ALDBBase(ABC):
    """Represents a base class for a device All-Link database."""

    def __init__(self, address, version=ALDBVersion.V2, mem_addr=0x0FFF):
        """Instantiate the ALL-Link Database object."""
        self._records = {}
        self._status = ALDBStatus.EMPTY
        self._version = version

        self._address = Address(address)
        self._mem_addr = mem_addr
        self._cb_aldb_loaded = None
        self._read_manager = None
        self._dirty_records = []

    def __len__(self):
        """Return the number of devices in the ALDB."""
        return len(self._records)

    def __iter__(self):
        """Iterate through each ALDB device record."""
        for key in sorted(self._records, reverse=True):
            yield key

    def __getitem__(self, mem_addr):
        """Fetch a device from the ALDB."""
        return self._records.get(mem_addr)

    @abstractmethod
    def __setitem__(self, mem_addr, record):
        """Add or Update a device in the ALDB."""

    def __repr__(self):
        """Human representation of a device from the ALDB."""
        attrs = vars(self)
        return ", ".join("%s: %r" % item for item in attrs.items())

    @property
    def address(self) -> Address:
        """Returnt the status of the device."""
        return self._address

    @property
    def status(self) -> ALDBStatus:
        """Return the ALDB load status."""
        return self._status

    @property
    def version(self) -> ALDBVersion:
        """Return the ALDB version."""
        return self._version

    @property
    def first_mem_addr(self):
        """Return the expected memory address of the first record."""
        return self._mem_addr

    @property
    def high_water_mark_mem_addr(self):
        """Return the High Water Mark record memory address."""
        for mem_addr in self._records:
            record = self._records[mem_addr]
            if record.is_high_water_mark:
                return record.mem_addr
        return None

    @property
    def is_loaded(self) -> bool:
        """Test if the ALDB is loaded."""
        return self._status == ALDBStatus.LOADED

    def get(self, mem_addr, default=None):
        """Get the record at address 'mem_addr'."""
        return self._records.get(mem_addr, default)

    def get_responders(self, group):
        """Return all responders to this device for a group."""
        for mem_addr in self._records:
            rec = self._records[mem_addr]
            if rec.is_controller and rec.group == group:
                yield rec.target

    @abstractmethod
    async def async_load(self, *args, **kwargs):
        """Load the All-Link Database."""

    def load_saved_records(self, status: ALDBStatus, records: [ALDBRecord]):
        """Load All-Link records from a dictionary of saved records."""
        if isinstance(status, ALDBStatus):
            self._status = status
        else:
            self._status = ALDBStatus(status)
        for mem_addr in records:
            record = records[mem_addr]
            self._notify_change(record, force_delete=True)
            self._records[mem_addr] = record
            self._notify_change(record)
        if self.is_loaded and self._records:
            keys = list(self._records.keys())
            keys.sort(reverse=True)
            self._mem_addr = keys[0]

    def _notify_change(self, record, force_delete=False):
        from .. import devices

        target = record.target
        group = record.group
        is_in_use = True if force_delete else record.is_in_use
        if group == 0 or target == devices.modem.address:
            return
        if record.is_controller and is_in_use:
            self._send_change(
                DEVICE_LINK_CONTROLLER_CREATED, self._address, target, group
            )
        elif record.is_controller and not is_in_use:
            self._send_change(
                DEVICE_LINK_CONTROLLER_REMOVED, self._address, target, group
            )
        elif not record.is_controller and is_in_use:
            self._send_change(
                DEVICE_LINK_RESPONDER_CREATED, self._address, target, group
            )
        else:
            self._send_change(
                DEVICE_LINK_RESPONDER_REMOVED, self._address, target, group
            )

    @classmethod
    def _send_change(cls, topic, controller, responder, group):
        pub.sendMessage(topic, controller=controller, responder=responder, group=group)


class ALDB(ALDBBase):
    """All-Link Database for a device."""

    def __init__(self, address, version=ALDBVersion.V2, mem_addr=0x0FFF):
        """Init the ALDB class."""
        from ..managers.aldb_read_manager import ALDBReadManager
        from ..managers.aldb_write_manager import ALDBWriteManager

        super().__init__(address=address, version=version, mem_addr=mem_addr)
        self._read_manager = ALDBReadManager(self)
        self._write_manager = ALDBWriteManager(self)

    def __setitem__(self, mem_addr, record):
        """Add or Update a device in the ALDB."""
        if not isinstance(record, ALDBRecord):
            raise ValueError

        self._dirty_records.append(record)

    def load(self, refresh=False, callback: Callable = None):
        """Load the ALDB calling the callback when done."""
        self._cb_aldb_loaded = callback
        asyncio.ensure_future(self.async_load(refresh))

    # pylint: disable=arguments-differ
    async def async_load(
        self,
        mem_addr: int = 0x00,
        num_recs: int = 0x00,
        refresh: bool = False,
        callback: Callable = None,
    ):
        """Load the All-Link Database."""
        _LOGGER.debug("Loading the ALDB async")
        self._status = ALDBStatus.LOADING
        if refresh:
            for maddr in self._records:
                self._notify_change(self._records[maddr], force_delete=True)
            self._records = {}
        async for rec in self._read_manager.async_read(
            mem_addr=mem_addr, num_recs=num_recs
        ):
            if self._records.get(rec.mem_addr):
                self._notify_change(self._records[rec.mem_addr], force_delete=True)
            self._records[rec.mem_addr] = rec
            self._notify_change(rec)
        self._set_load_status()
        if callback:
            callback()
        return self._status

    def write(self):
        """Write modified records to the device."""
        if self.is_loaded:
            asyncio.ensure_future(self.async_write())

    async def async_write(self):
        """Write modified records to the device.

        Returns a tuple of (completed, failed) record counts.
        """
        from ..handlers import ResponseStatus

        completed = []
        failed = []
        while self._dirty_records:
            record = self._dirty_records.pop()
            if record.mem_addr == 0x0000:
                mem_addr = self._existing_link(
                    record.is_controller, record.group, record.target
                )
                if mem_addr is None:
                    mem_addr = self._get_next_mem_addr()
                record.mem_addr = mem_addr
            # We assume a direct ACK is a confirmation of write.
            # Should we re-read to ensure it is correct.
            response = ResponseStatus.UNSENT
            retries = 0
            while response != ResponseStatus.SUCCESS and retries < 3:
                response = await self._write_manager.async_write(record)
                retries += 1
            if response == ResponseStatus.SUCCESS:
                completed.append(record)
            else:
                failed.append(record)
        for record in failed:
            self._dirty_records.append(record)
        return len(completed), len(self._dirty_records)

    def add(
        self,
        group: int,
        target: Address,
        controller: bool = False,
        data1: int = 0x00,
        data2: int = 0x00,
        data3: int = 0x00,
        bit5: int = True,
        bit4: int = False,
    ):
        """Add an All-Link record.

        This method does not write to the device. To write modifications to the device
        use `write` or `async_write`.
        """
        mem_addr = 0x0000

        rec = ALDBRecord(
            memory=mem_addr,
            in_use=True,
            controller=controller,
            high_water_mark=False,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            bit5=bit5,
            bit4=bit4,
        )
        self._dirty_records.append(rec)

    def remove(self, mem_addr: int):
        """Remove an All-Link record."""
        rec = self._records.get(mem_addr)
        if not rec:
            raise IndexError("Memory location not found.")
        new_rec = ALDBRecord(
            memory=rec.mem_addr,
            in_use=False,
            controller=rec.is_controller,
            high_water_mark=rec.is_high_water_mark,
            group=rec.group,
            target=rec.target,
            data1=rec.data1,
            data2=rec.data2,
            data3=rec.data3,
            bit5=rec.is_bit5_set,
            bit4=rec.is_bit4_set,
        )
        self._dirty_records.append(new_rec)

    def _existing_link(self, is_controller, group, address):
        """Test if a link exists in a device ALDB."""
        for mem_addr in self._records:
            rec = self._records[mem_addr]
            if (
                rec.is_controller == is_controller
                and rec.target == address
                and rec.group == group
            ):
                return rec.mem_addr
        return None

    def _get_next_mem_addr(self):
        """Return the next memory slot available."""
        if not self.is_loaded:
            return None

        last_record = None
        for mem_addr in self._records:
            record = self._records[mem_addr]
            last_record = record
            if record.is_high_water_mark or not record.is_in_use:
                return record.mem_addr

        return last_record.mem_addr - 8

    def calc_load_status(self):
        """Test if the ALDB is fully loaded."""
        has_first = False
        has_last = False
        has_all = False
        last_addr = 0x0000
        first = True
        for mem_addr in sorted(self._records, reverse=True):
            if first:
                _LOGGER.debug("First Addr: 0x%4x", mem_addr)
            if mem_addr == self._mem_addr:
                has_first = True
            if self._records[mem_addr].is_high_water_mark:
                has_last = True
            if last_addr != 0x0000:
                has_all = (last_addr - mem_addr) == 8
            last_addr = mem_addr
        _LOGGER.debug("Has First is %s", has_first)
        _LOGGER.debug("Has Last is %s", has_last)
        _LOGGER.debug("Has All is %s", has_all)
        return has_first and has_all and has_last

    def _set_load_status(self):
        _LOGGER.debug("Setting the load status")
        if self.calc_load_status():
            self._status = ALDBStatus.LOADED
        elif self._records:
            self._status = ALDBStatus.PARTIAL
        else:
            self._status = ALDBStatus.EMPTY



"""ALDB Status representation."""
from enum import Enum


class ALDBStatus(Enum):
    """All-Link Database load status."""

    EMPTY = 0
    LOADING = 1
    LOADED = 2
    FAILED = 3
    PARTIAL = 4



"""Device state definitions."""
from abc import ABC, abstractmethod
from ..address import Address
from ..subscriber_base import SubscriberBase


DIMMABLE_LIGHT = "dimmable_light"
DIMMABLE_FAN = "dimmable_fan"
DIMMABLE_OUTLET = "dimmable_outlet"
ON_OFF_SWITCH = "on_off_switch"
ON_OFF_OUTLET_TOP = "on_off_outlet_top"
ON_OFF_OUTLET_BOTTOM = "on_off_outlet_bottom"
LOW_BATTERY = "low_battery"
OPEN_CLOSE_SENSOR = "open_close_sensor"
LIGHT_SENSOR = "light_sensor"
LEAK_SENSOR = "leak_sensor"
DOOR_SENSOR = "door_sensor"
MOTION_SENSOR = "motion_sensor"
SMOKE_SENSOR = "smoke_sensor"
CO_SENSOR = "co_sensor"
TEST_SENSOR = "test_sensor"
NEW_SENSOR = "new_sensor"
HEARTBEAT = "heartbeat"
SENSOR_MALFUNCTION = "sensor_malfunction"
COVER = "cover"

ON_OFF_SWITCH_MAIN = "{}_{}".format(ON_OFF_SWITCH, "main")
ON_OFF_SWITCH_A = "{}_{}".format(ON_OFF_SWITCH, "a")
ON_OFF_SWITCH_B = "{}_{}".format(ON_OFF_SWITCH, "b")
ON_OFF_SWITCH_C = "{}_{}".format(ON_OFF_SWITCH, "c")
ON_OFF_SWITCH_D = "{}_{}".format(ON_OFF_SWITCH, "d")
ON_OFF_SWITCH_E = "{}_{}".format(ON_OFF_SWITCH, "e")
ON_OFF_SWITCH_F = "{}_{}".format(ON_OFF_SWITCH, "f")
ON_OFF_SWITCH_G = "{}_{}".format(ON_OFF_SWITCH, "g")
ON_OFF_SWITCH_H = "{}_{}".format(ON_OFF_SWITCH, "h")

DIMMABLE_LIGHT_MAIN = "{}_{}".format(DIMMABLE_LIGHT, "main")
# DIMMABLE_LIGHT_A = '{}_{}'.format(DIMMABLE_LIGHT, 'a')
# DIMMABLE_LIGHT_B = '{}_{}'.format(DIMMABLE_LIGHT, 'b')
# DIMMABLE_LIGHT_C = '{}_{}'.format(DIMMABLE_LIGHT, 'c')
# DIMMABLE_LIGHT_D = '{}_{}'.format(DIMMABLE_LIGHT, 'd')
# DIMMABLE_LIGHT_E = '{}_{}'.format(DIMMABLE_LIGHT, 'e')
# DIMMABLE_LIGHT_F = '{}_{}'.format(DIMMABLE_LIGHT, 'f')
# DIMMABLE_LIGHT_G = '{}_{}'.format(DIMMABLE_LIGHT, 'g')
# DIMMABLE_LIGHT_H = '{}_{}'.format(DIMMABLE_LIGHT, 'h')


class StateBase(SubscriberBase):
    """Device state base class."""

    __meta__ = ABC

    def __init__(
        self, name: str, address: Address, group=0, default=None, value_type: type = int
    ):
        """Init the StateBase class."""
        self._address = Address(address)
        topic = "state_{}_{}_{}".format(self._address.id, name, group)
        super().__init__(subscriber_topic=topic)
        self._name = name
        self._group = group
        self._value = int(default) if default is not None else None
        self._type = value_type

    @property
    def name(self):
        """Return the name of the state."""
        return self._name

    @property
    def value(self):
        """Return the value of the state."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the value of the state."""
        try:
            self._value = self._type(value) if value is not None else None
        except TypeError:
            raise TypeError(
                "Error setting value of State {}: "
                "Must be of type {}".format(self._name, self._type.__name__)
            )
        else:
            self._call_subscribers(
                name=self._name,
                address=self._address.id,
                value=self._value,
                group=self._group,
            )

    @abstractmethod
    def set_value(self, **kwargs):
        """Set the value of the state from a Handler."""



"""Sensor/Actuator devices."""
from ..events import CLOSE_EVENT, OPEN_EVENT, Event
from ..managers.on_level_manager import OnLevelManager
from ..states import OPEN_CLOSE_SENSOR
from ..states.open_close import NormallyOpen
from .on_off_responder_base import OnOffResponderBase
from ..handlers.to_device.status_request import StatusRequestCommand, STATUS_REQUEST

ON_LEVEL_MANAGER = "on_level_manager"
SENSOR_GROUP = 2


class SensorsActuators(OnOffResponderBase):
    """Sensor Actuator device."""


class SensorsActuators_IOLink(SensorsActuators):
    """I/O Link device."""

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        if self._managers.get(SENSOR_GROUP) is None:
            self._managers[SENSOR_GROUP] = {}
        self._managers[SENSOR_GROUP][ON_LEVEL_MANAGER] = OnLevelManager(
            self._address, SENSOR_GROUP
        )
        if self._handlers.get(SENSOR_GROUP) is None:
            self._handlers[SENSOR_GROUP] = {}
        self._handlers[SENSOR_GROUP][STATUS_REQUEST] = StatusRequestCommand(
            self.address, 1
        )

    def status(self, group=None):
        """Get the status of the relay and/or the sensor."""
        if group in [1, None]:
            super().status()
        if group == [2, None]:
            self.sensor_status()

    async def async_status(self, group=None):
        """Get the status of the relay and/or the sensor."""
        from ..utils import multiple_status

        response_1 = None
        response_2 = None
        if group in [1, None]:
            response_1 = await super().async_status()
        if group == [2, None]:
            response_2 = await self.async_sensor_status()
        return multiple_status(response_1, response_2)

    def relay_status(self):
        """Get the status of the relay switch."""
        super().status()

    async def async_relay_status(self):
        """Get the status of the relay switch."""
        return await super().async_status()

    def sensor_status(self):
        """Get the status of the sensor."""
        self._handlers[SENSOR_GROUP][STATUS_REQUEST].send()

    async def async_sensor_status(self):
        """Get the status of the sensor."""
        return await self._handlers[SENSOR_GROUP][STATUS_REQUEST].async_send()

    def _register_states(self):
        super()._register_states()
        # Off is a Open state and On is an Closed state
        self._states[SENSOR_GROUP] = NormallyOpen(
            OPEN_CLOSE_SENSOR, self._address, SENSOR_GROUP
        )

    def _register_events(self):
        super()._register_events()
        if self._events.get(SENSOR_GROUP) is None:
            self._events[SENSOR_GROUP] = {}
        self._events[SENSOR_GROUP][OPEN_EVENT] = Event(
            OPEN_EVENT, self._address, SENSOR_GROUP
        )
        self._events[SENSOR_GROUP][CLOSE_EVENT] = Event(
            CLOSE_EVENT, self._address, SENSOR_GROUP
        )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._managers[SENSOR_GROUP][ON_LEVEL_MANAGER].subscribe(
            self._states[SENSOR_GROUP].set_value
        )
        self._managers[SENSOR_GROUP][ON_LEVEL_MANAGER].subscribe_off(
            self._events[SENSOR_GROUP][OPEN_EVENT].trigger
        )
        self._managers[SENSOR_GROUP][ON_LEVEL_MANAGER].subscribe_on(
            self._events[SENSOR_GROUP][CLOSE_EVENT].trigger
        )

    def _register_operating_flags(self):
        from ..operating_flag import (
            PROGRAM_LOCK_ON,
            LED_BLINK_ON_TX_ON,
            RELAY_ON_SENSE_ON,
            MOMENTARY_MODE_ON,
            MOMENTARY_ON_OFF_TRIGGER,
            MOMENTARY_FOLLOW_SENSE,
            X10_OFF,
        )
        from ..extended_property import X10_HOUSE, X10_UNIT, DELAY, PRESCALER

        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RELAY_ON_SENSE_ON, 0, 2, 4, 5)
        self._add_operating_flag(MOMENTARY_MODE_ON, 0, 3, 6, 7)
        self._add_operating_flag(MOMENTARY_ON_OFF_TRIGGER, 0, 4, 9, 8)
        self._add_operating_flag(X10_OFF, 0, 5, 10, 11)
        self._add_operating_flag(MOMENTARY_FOLLOW_SENSE, 0, 6, 12, 13)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 7, 14, 15)

        self._add_property(PRESCALER, 2, 7)
        self._add_property(DELAY, 3, 6)
        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)



"""Base device object."""

import asyncio
from functools import partial
import logging

from .. import pub
from ..handlers.to_device.extended_set import ExtendedSetCommand


_LOGGER = logging.getLogger(__name__)
TIMEOUT = 2


# pylint: disable=no-member
class BatteryDeviceBase:
    """Base class for battery operated devices."""

    def __init__(
        self, address, cat, subcat, firmware=0x00, description="", model="", **kwargs
    ):
        """Init the DeviceBattery class."""
        from ..aldb.aldb_battery import ALDBBattery

        # The super class is the non-battery base class such as OnOffControllerBase
        super(BatteryDeviceBase, self).__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=0x00,
            description="",
            model="",
            **kwargs
        )
        self._commands_queue = asyncio.Queue()
        self._aldb = ALDBBattery(address=address, run_command=self._run_on_wake)
        self._last_run = None
        pub.subscribe(self._device_awake, self._address.id)
        asyncio.ensure_future(self.async_keep_awake())

    def _run_on_wake(self, command, **kwargs):
        cmd = partial(command, **kwargs)
        self._commands_queue.put_nowait(cmd)

    async def async_read_op_flags(self):
        """Read the device operating flags."""
        self._run_on_wake(super(BatteryDeviceBase, self).async_read_op_flags)

    async def async_write_op_flags(self):
        """Write the operating flags to the device."""
        self._run_on_wake(super(BatteryDeviceBase, self).async_write_op_flags)

    async def async_read_ext_properties(self):
        """Get the device extended properties."""
        self._run_on_wake(super(BatteryDeviceBase, self).async_read_ext_properties)

    async def async_keep_awake(self, awake_time=0xFF):
        """Keep the device awake to ensure commands are heard."""
        cmd = ExtendedSetCommand(self._address, data1=0, data2=0x04)
        return await cmd.async_send(data3=awake_time)

    def _device_awake(self, topic=pub.AUTO_TOPIC, **kwargs):
        """Execute the commands that were requested while sleeping."""
        if self._last_run is None or self._last_run.done():
            self._last_run = asyncio.ensure_future(self._run_commands())

    async def _run_commands(self):
        from inspect import iscoroutinefunction, iscoroutine

        await self.async_keep_awake()
        await asyncio.sleep(2)
        try:
            command = await asyncio.wait_for(self._commands_queue.get(), TIMEOUT)
            if isinstance(command, partial):
                if iscoroutine(command.func) or iscoroutinefunction(command.func):
                    await command()
                else:
                    command()
            elif iscoroutine(command) or iscoroutinefunction(command):
                await command()
            else:
                command()
        except asyncio.TimeoutError:
            pass



"""Fan On Level state."""
from . import StateBase
from ..constants import FanSpeed
from ..address import Address


class FanOnLevel(StateBase):
    """On / Off state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: FanSpeed = None
    ):
        """Init the FanOnLevel class."""
        super().__init__(name, address, group, default, value_type=FanSpeed)

    # pylint: disable=arguments-differ
    def set_value(self, on_level):
        """Set the value of the state from the handlers."""
        fan_speed = FanSpeed.OFF
        if int(FanSpeed.OFF) < on_level <= int(FanSpeed.LOW):
            fan_speed = FanSpeed.LOW
        if int(FanSpeed.LOW) < on_level <= int(FanSpeed.MEDIUM):
            fan_speed = FanSpeed.MEDIUM
        if int(FanSpeed.MEDIUM) < on_level <= int(FanSpeed.HIGH):
            fan_speed = FanSpeed.LOW
        self.value = fan_speed



"""ALDB Read Manager."""
import asyncio
import logging

from async_generator import async_generator, yield_

from ..aldb.aldb_record import ALDBRecord

RETRIES_ALL_MAX = 5
RETRIES_ONE_MAX = 20
RETRIES_WRITE_MAX = 5
TIMER = 5
TIMER_INCREMENT = 3
_LOGGER = logging.getLogger(__name__)
READ_ALL = 1
READ_ONE = 2
WRITE = 3
CANCEL = 0

IM_NOT_IN_DEVICE_ALDB = 0xFF
CHECKSUM_ERROR = 0xFD
ILLEGAL_VALUE_IN_COMMAND = 0xFB


class ALDBReadManager:
    """ALDB Read Manager."""

    def __init__(self, aldb, mem_addr: int = 0x00, num_recs: int = 0):  # : ALDB):
        """Init the ALDBReadManager class."""
        from ..handlers.to_device.read_aldb import ReadALDBCommandHandler
        from ..handlers.from_device.receive_aldb_record import ReceiveALDBRecordHandler

        self._aldb = aldb
        self._mem_addr = mem_addr
        self._num_recs = num_recs

        self._records = asyncio.Queue()
        self._retries_all = 0
        self._retries_one = 0
        self._retries_write = 0
        self._last_command = None
        self._last_mem_addr = 0
        self._read_handler = ReadALDBCommandHandler(self._aldb.address)
        self._record_handler = ReceiveALDBRecordHandler(self._aldb.address)
        self._read_handler.subscribe(self._receive_direct_ack)
        self._record_handler.subscribe(self._receive_record)
        self._timer_lock = asyncio.Lock()

    @async_generator
    async def async_read(self, mem_addr: int = 0x00, num_recs: int = 0):
        """Start the reading process to enable iteration."""
        if self._mem_addr == 0x0000 and self._num_recs == 0:
            self._last_command = READ_ALL
        else:
            self._last_command = READ_ONE
        self._last_mem_addr = self._mem_addr
        await self._async_read(mem_addr=self._mem_addr, num_recs=self._num_recs)
        while True:
            record = await self._records.get()
            if record is None:
                break
            await yield_(record)

    async def _async_read(self, mem_addr: int = 0x00, num_recs: int = 0):
        """Perform the device read function."""
        if self._last_command == READ_ALL and self._retries_all < RETRIES_ALL_MAX:
            retries = self._retries_all
        else:
            retries = self._retries_one
        _LOGGER.debug("Attempting to read %x", mem_addr)
        # TODO check for success or failure
        await self._read_handler.async_send(mem_addr=mem_addr, num_recs=num_recs)
        timer = TIMER + retries * TIMER_INCREMENT
        asyncio.ensure_future(self._timer(timer, mem_addr, num_recs))

    def _receive_direct_ack(self, ack_response):
        """Receive the response from the direct ACK."""
        if ack_response in [
            IM_NOT_IN_DEVICE_ALDB,
            CHECKSUM_ERROR,
            ILLEGAL_VALUE_IN_COMMAND,
        ]:
            _LOGGER.error("ALDB Load error: 0x%02x", ack_response)
            self._last_command = CANCEL
            self._records.put_nowait(None)

    def _receive_record(
        self,
        memory,
        controller,
        group,
        target,
        data1,
        data2,
        data3,
        in_use,
        high_water_mark,
        bit5,
        bit4,
    ):
        """Receive an ALDB record."""
        record = ALDBRecord(
            memory=memory,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
        self._records.put_nowait(record)
        asyncio.ensure_future(self._release_timer())

    async def _release_timer(self):
        """Release the lock that tells `_timer` to quit."""
        if self._last_command == READ_ALL and self._retries_all < RETRIES_ALL_MAX:
            # Wait to see if more messages come in.
            await asyncio.sleep(5)
        if self._timer_lock.locked():
            self._timer_lock.release()

    async def _timer(self, timer, mem_addr, num_recs):
        """Set a timer to confirm if the last get command completed."""
        if self._timer_lock.locked():
            self._timer_lock.release()
        await self._timer_lock.acquire()
        try:
            await asyncio.wait_for(self._timer_lock.acquire(), timer)
        except asyncio.TimeoutError:
            pass
        if self._last_command == READ_ALL:
            self._manage_get_all_cmd(mem_addr, num_recs)
        elif self._last_command == READ_ONE:
            self._manage_get_one_cmd(mem_addr, num_recs)

    def _manage_get_all_cmd(self, mem_addr, num_recs):
        """Manage the READ_ALL command process."""
        _LOGGER.debug("In _manage_get_all_cmd")
        if self._aldb.calc_load_status():
            # The ALDB is fully loaded so stop
            self._records.put_nowait(None)
            return

        if self._retries_all < RETRIES_ALL_MAX:
            # Attempt to read all records again
            asyncio.ensure_future(self._async_read(0x0000, 0))
            self._retries_all += 1
            _LOGGER.info("Retry reading all records %d times", self._retries_all)
        else:
            # Read the next missing record
            next_mem_addr = self._next_missing_record()
            if next_mem_addr is None:
                # No records are missing so quit
                self._records.put_nowait(None)
                return
            if next_mem_addr == self._last_mem_addr:
                # We are still trying to get the same record as the last read
                if self._retries_one < RETRIES_ONE_MAX:
                    asyncio.ensure_future(self._async_read(next_mem_addr, 1))
                    self._retries_one += 1
                    _LOGGER.info(
                        "Retry reading 0x%04x record %d times",
                        next_mem_addr,
                        self._retries_one,
                    )
                else:
                    # Tried to read the same record max times so quit
                    self._records.put_nowait(None)
                    return
            else:
                # Reading a different record than the last attempt so reset
                # the retry count
                self._last_mem_addr = next_mem_addr
                self._retries_one = 0
                asyncio.ensure_future(self._async_read(next_mem_addr, 1))

    def _manage_get_one_cmd(self, mem_addr, num_recs):
        """Manage the READ_ONE command process."""
        if self._aldb.get(mem_addr):
            self._records.put_nowait(None)
            return
        if self._retries_one < RETRIES_ONE_MAX:
            asyncio.ensure_future(
                self._async_read(mem_addr=mem_addr, num_recs=num_recs)
            )
            self._retries_one += 1
            _LOGGER.info(
                "Retry reading 0x%04x record %d times", mem_addr, self._retries_one
            )
        else:
            # Trigger aldb.loaded but this will check the load status.
            self._records.put_nowait(None)
            return

    def _next_missing_record(self):
        last_addr = 0
        if not self._has_first_record():
            if self._last_mem_addr == 0x0000 and self._retries_one < RETRIES_ONE_MAX:
                return 0x0000
            return self._aldb.first_mem_addr
        for mem_addr in self._aldb:
            rec = self._aldb[mem_addr]
            if rec.is_high_water_mark:
                return None
            if last_addr != 0:
                if not last_addr - 8 == mem_addr:
                    return last_addr - 8
            last_addr = mem_addr
        return last_addr - 8

    def _has_first_record(self):
        """Test if the first record is loaded."""
        for mem_addr in self._aldb:
            if mem_addr in [self._aldb.first_mem_addr, 0x0FFF]:
                _LOGGER.debug("Found First record: 0x%04x", mem_addr)
                return True
        return False



"""Manages links between devices to identify device state of responders."""
import asyncio
from .. import pub
from ..address import Address
from ..topics import (
    DEVICE_LINK_CONTROLLER_CREATED,
    DEVICE_LINK_RESPONDER_CREATED,
    DEVICE_LINK_CONTROLLER_REMOVED,
    DEVICE_LINK_RESPONDER_REMOVED,
)
from ..constants import MessageFlagType


def _controller_group_topic(responder, group):
    return "{}.{}".format(responder.id, group)


def _topic_to_addr_group(topic):
    elements = topic.name.split(".")
    try:
        address = Address(elements[0])
        group = int(elements[1])
        msg_type = getattr(MessageFlagType, elements[-1].upper())
    except [KeyError, TypeError]:
        return None, None
    return address, group, msg_type


class DeviceLinkManager:
    """Manages links between devices to identify state of responders."""

    def __init__(self):
        """Init the DeviceLinkManager class."""
        pub.subscribe(self._controller_link_created, DEVICE_LINK_CONTROLLER_CREATED)
        pub.subscribe(self._responder_link_created, DEVICE_LINK_RESPONDER_CREATED)
        pub.subscribe(self._controller_link_removed, DEVICE_LINK_CONTROLLER_REMOVED)
        pub.subscribe(self._responder_link_removed, DEVICE_LINK_RESPONDER_REMOVED)
        self._responders = {}

    def _controller_link_created(self, controller, responder, group):
        controller_group = _controller_group_topic(controller, group)
        try:
            topic = pub.getDefaultTopicMgr().getTopic(controller_group)
        except pub.TopicNameError:
            topic = None
        if not topic or (
            not pub.isSubscribed(self._check_responder, controller_group)
            and not pub.isSubscribed(self._check_controller, controller_group)
        ):
            pub.subscribe(self._check_controller, controller_group)

    def _responder_link_created(self, controller, responder, group):
        controller_group = _controller_group_topic(controller, group)
        try:
            topic = pub.getDefaultTopicMgr().getTopic(controller_group)
        except pub.TopicNameError:
            topic = None
        if topic and pub.isSubscribed(self._check_controller, controller_group):
            pub.unsubscribe(self._check_controller, controller_group)
        if not topic or not pub.isSubscribed(self._check_responder, controller_group):
            pub.subscribe(self._check_responder, controller_group)

        if self._responders.get(controller) is None:
            self._responders[controller] = {}
        if self._responders[controller].get(group) is None:
            self._responders[controller][group] = []
        self._responders[controller][group].append(responder)

    def _controller_link_removed(self, controller, responder, group):
        controller_group = _controller_group_topic(controller, group)
        try:
            topic = pub.getDefaultTopicMgr().getTopic(controller_group)
        except pub.TopicNameError:
            topic = None
        if topic and pub.isSubscribed(self._check_controller, controller_group):
            pub.subscribe(self._check_controller, controller_group)

    def _responder_link_removed(self, controller, responder, group):
        controller_group = _controller_group_topic(controller, group)
        try:
            topic = pub.getDefaultTopicMgr().getTopic(controller_group)
        except pub.TopicNameError:
            topic = None
        if topic and pub.isSubscribed(self._check_responder, controller_group):
            pub.subscribe(self._check_responder, controller_group)

        groups = self._responders.get(controller)
        if groups and groups.get(group):
            groups[group].remove(responder)

    def _check_responder(self, on_level, topic=pub.AUTO_TOPIC):
        from .. import devices

        controller, group, msg_type = _topic_to_addr_group(topic)
        if msg_type != MessageFlagType.ALL_LINK_BROADCAST:
            return
        groups = self._responders.get(controller)
        if groups:
            responder = groups.get(group)
            if group:
                devices[responder].states[group].set_value(on_level)
                asyncio.ensure_future(devices[responder].async_status())

    @classmethod
    def _check_controller(cls, on_level, topic=pub.AUTO_TOPIC):
        from .. import devices

        controller, group, msg_type = _topic_to_addr_group(topic)
        if msg_type != MessageFlagType.ALL_LINK_BROADCAST:
            return
        if controller is None or group is None:
            return
        for responder in devices[controller].aldb.get_responders(group):
            asyncio.ensure_future(devices[responder].async_status())



"""Dimmable Lighting Control Devices (CATEGORY 0x01)."""
from ..constants import FanSpeed
from ..handlers.to_device.set_leds import SetLedsCommandHandler
from ..handlers.to_device.status_request import StatusRequestCommand

# from ..handlers.to_device.trigger_scene_on import TriggerSceneOnCommandHandler
# from ..handlers.to_device.trigger_scene_off import TriggerSceneOffCommandHandler
from ..states import (
    DIMMABLE_FAN,
    DIMMABLE_LIGHT,
    DIMMABLE_LIGHT_MAIN,
    ON_OFF_SWITCH_A,
    ON_OFF_SWITCH_B,
    ON_OFF_SWITCH_C,
    ON_OFF_SWITCH_D,
    ON_OFF_SWITCH_E,
    ON_OFF_SWITCH_F,
    ON_OFF_SWITCH_G,
    ON_OFF_SWITCH_H,
    DIMMABLE_OUTLET,
)
from ..states.on_off import OnOff
from .commands import (
    OFF_COMMAND,
    OFF_FAST_COMMAND,
    ON_COMMAND,
    ON_FAST_COMMAND,
    SET_LEDS_COMMAND,
    STATUS_COMMAND_FAN,
    # TRIGGER_SCENE_ON_COMMAND,
    # TRIGGER_SCENE_OFF_COMMAND,
)
from .variable_responder_base import VariableResponderBase


class DimmableLightingControl(VariableResponderBase):
    """Dimmable Lighting Control Device."""


class DimmableLightingControl_LampLinc(DimmableLightingControl):
    """LampLinc based dimmable lights."""

    def _register_operating_flags(self):
        from ..operating_flag import (
            PROGRAM_LOCK_ON,
            LED_BLINK_ON_TX_ON,
            RESUME_DIM_ON,
            LED_ON,
            LOAD_SENSE_ON,
        )
        from ..extended_property import (
            LED_DIMMING,
            ON_LEVEL,
            X10_HOUSE,
            X10_UNIT,
            RAMP_RATE,
        )

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9)
        self._add_operating_flag(LOAD_SENSE_ON, 0, 5, 0x0A, 0x0B)

        self._add_property(LED_DIMMING, 3, 3)
        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)
        self._add_property(RAMP_RATE, 7, 5)
        self._add_property(ON_LEVEL, 8, 6)


class DimmableLightingControl_SwitchLinc(DimmableLightingControl):
    """SwichLinc based dimmable lights."""

    def _register_operating_flags(self):
        from ..operating_flag import (
            PROGRAM_LOCK_ON,
            LED_BLINK_ON_TX_ON,
            RESUME_DIM_ON,
            LED_ON,
            KEY_BEEP_ON,
            LED_BLINK_ON_ERROR_ON,
        )
        from ..extended_property import (
            LED_DIMMING,
            ON_LEVEL,
            X10_HOUSE,
            X10_UNIT,
            RAMP_RATE,
        )

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)
        self._add_operating_flag(LED_BLINK_ON_ERROR_ON, 5, 2, 0x14, 0x15)

        self._add_property(LED_DIMMING, 3, 3)
        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)
        self._add_property(RAMP_RATE, 7, 5)
        self._add_property(ON_LEVEL, 8, 6)


class DimmableLightingControl_InLineLinc(DimmableLightingControl_SwitchLinc):
    """InLineLinc based dimmable lights."""


class DimmableLightingControl_OutletLinc(DimmableLightingControl):
    """OutletLinc based dimmable lights."""

    def __init__(self, address, cat, subcat, firmware=0, description="", model=""):
        """Init the DimmableLightingControl_OutletLinc class."""
        buttons = {1: DIMMABLE_OUTLET}
        super().__init__(
            address,
            cat,
            subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )

    def _register_operating_flags(self):
        from ..operating_flag import PROGRAM_LOCK_ON, LED_BLINK_ON_TX_ON, LED_ON
        from ..extended_property import X10_HOUSE, X10_UNIT

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9)

        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)


class DimmableLightingControl_DinRail(DimmableLightingControl):
    """DINRail based dimmable lights."""

    def _register_operating_flags(self):
        from ..operating_flag import (
            PROGRAM_LOCK_ON,
            LED_BLINK_ON_TX_ON,
            LED_ON,
            KEY_BEEP_ON,
        )
        from ..extended_property import (
            LED_DIMMING,
            ON_LEVEL,
            X10_HOUSE,
            X10_UNIT,
            RAMP_RATE,
        )

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)

        self._add_property(LED_DIMMING, 3, 3)
        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)
        self._add_property(RAMP_RATE, 7, 5)
        self._add_property(ON_LEVEL, 8, 6)


class DimmableLightingControl_FanLinc(DimmableLightingControl):
    """FanLinc model 2475F Dimmable Lighting Control.

    Device Class 0x01 subcat 0x2e

    """

    def __init__(self, address, cat, subcat, firmware=0, description="", model=""):
        """Init the DimmableLightingControl_FanLinc class."""
        buttons = {1: DIMMABLE_LIGHT, 2: DIMMABLE_FAN}
        super().__init__(
            address,
            cat,
            subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )

    def fan_on(self, on_level: FanSpeed = FanSpeed.HIGH, fast=False):
        """Turn on the fan.

        Parameters:
            on_level: Default FanSpeed.HIGH (full on). Set the fan on level.
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        """
        command = ON_FAST_COMMAND if fast else ON_COMMAND
        command_group = "{}_2".format(command)
        fan_speed = int(on_level)
        # The send command converts to int
        self._handlers[command_group].send(on_level=fan_speed)

    async def async_fan_on(self, on_level: FanSpeed = FanSpeed.HIGH, fast=False):
        """Turn on the device.

        Parameters:
            on_level: Default FanSpeed.HIGH (full on). Set the fan on level.
            fast: Default False. If True, bypass fan ramp rate otherwise
            turn on at the ramp rate.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        command = ON_FAST_COMMAND if fast else ON_COMMAND
        command_group = "{}_2".format(command)
        fan_speed = int(on_level)
        # The send command converts to int
        return await self._handlers[command_group].async_send(on_level=fan_speed)

    def fan_off(self, fast=False):
        """Turn off the device.

        Parameters:
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        """
        command = OFF_FAST_COMMAND if fast else OFF_COMMAND
        command_group = "{}_2".format(command)
        self._handlers[command_group].send()

    async def async_fan_off(self, fast=False):
        """Turn off the device.

        Parameters:
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        command = OFF_FAST_COMMAND if fast else OFF_COMMAND
        command_group = "{}_2".format(command)
        return await self._handlers[command_group].async_send(group=2)

    def status(self):
        """Request the status of the light and the fan."""
        super().status()
        self.fan_status()

    async def async_status(self):
        """Request the status fo the light and the fan."""
        from ..handlers import ResponseStatus

        light_status = await super().async_status()
        fan_status = await self.async_fan_status()
        if light_status == fan_status == ResponseStatus.SUCCESS:
            return ResponseStatus.SUCCESS
        if (
            light_status == ResponseStatus.UNCLEAR
            or fan_status == ResponseStatus.UNCLEAR
        ):
            return ResponseStatus.UNCLEAR
        return ResponseStatus.FAILURE

    def light_status(self):
        """Request the status of the light."""
        super().status()

    async def async_light_status(self):
        """Request the status of the light."""
        return await super().async_status()

    def fan_status(self):
        """Request the status of the fan."""
        self._handlers[STATUS_COMMAND_FAN].send()

    async def async_fan_status(self):
        """Request the status of the fan.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        return await self._handlers[STATUS_COMMAND_FAN].async_send()

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[STATUS_COMMAND_FAN] = StatusRequestCommand(self._address, 3)

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._handlers[STATUS_COMMAND_FAN].subscribe(self._handle_fan_status)

    def _register_operating_flags(self):
        from ..operating_flag import (
            PROGRAM_LOCK_ON,
            LED_BLINK_ON_TX_ON,
            RESUME_DIM_ON,
            LED_OFF,
            KEY_BEEP_ON,
            RF_DISABLE_ON,
            POWERLINE_DISABLE_ON,
            DATABASE_DELTA,
            CRC_ERROR_COUNT,
            SIGNAL_TO_NOISE_FAILURE_COUNT,
            X10_OFF,
            LED_BLINK_ON_ERROR_ON,
            CLEANUP_REPORT_ON,
        )
        from ..extended_property import ON_LEVEL, X10_HOUSE, X10_UNIT, RAMP_RATE

        super()._register_operating_flags()

        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_OFF, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)
        self._add_operating_flag(RF_DISABLE_ON, 0, 6, 0x0C, 0x0D)
        self._add_operating_flag(POWERLINE_DISABLE_ON, 0, 7, 0x0E, 0x0F)

        self._add_operating_flag(DATABASE_DELTA, 1, None, None, None)
        self._add_operating_flag(CRC_ERROR_COUNT, 2, None, None, None)
        self._add_operating_flag(SIGNAL_TO_NOISE_FAILURE_COUNT, 3, None, None, None)

        self._add_operating_flag(X10_OFF, 5, 1, 0x12, 0x13)
        self._add_operating_flag(LED_BLINK_ON_ERROR_ON, 5, 2, 0x14, 0x15)
        self._add_operating_flag(CLEANUP_REPORT_ON, 5, 3, 0x16, 0x17)

        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)
        self._add_property(RAMP_RATE, 7, 5)
        self._add_property(ON_LEVEL, 8, 6)

    def _handle_fan_status(self, db_version, status):
        if int(status) == 0:
            self._states[2].set_value(FanSpeed.OFF)
        elif int(status) <= int(FanSpeed.LOW):
            self._states[2].set_value(FanSpeed.LOW)
        elif int(status) <= int(FanSpeed.MEDIUM):
            self._states[2].set_value(FanSpeed.MEDIUM)
        else:
            self._states[2].set_value(FanSpeed.HIGH)


# TODO setup operating flags for each KPL button
# TODO trigger scenes
class DimmableLightingControl_KeypadLinc(DimmableLightingControl):
    """KeypadLinc base class."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the GeneralController_MiniRemoteBase class."""
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )

    async def async_on(self, on_level: int = 0xFF, group: int = 0, fast: bool = False):
        """Turn on the button LED."""
        if group in [0, 1]:
            return await super().async_on(on_level=on_level, group=group, fast=fast)
        kwargs = self._change_led_status(led=group, is_on=True)
        return await self._handlers[SET_LEDS_COMMAND].async_send(**kwargs)

    async def async_off(self, group: int = 0, fast: bool = False):
        """Turn off the button LED."""
        if group in [0, 1]:
            return await super().async_off(group=group, fast=fast)
        kwargs = self._change_led_status(led=group, is_on=False)
        return await self._handlers[SET_LEDS_COMMAND].async_send(**kwargs)

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[SET_LEDS_COMMAND] = SetLedsCommandHandler(address=self.address)

    def _register_states(self):
        super()._register_states()
        for button in self._buttons:
            name = self._buttons[button]
            self._states[button] = OnOff(name=name, address=self._address, group=button)

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._handlers[SET_LEDS_COMMAND].subscribe(self._update_leds)

    def _change_led_status(self, led, is_on):
        leds = {}
        for curr_led in range(1, 9):
            var = "group{}".format(curr_led)
            leds[var] = is_on if curr_led == led else bool(self._states.get(curr_led))
        return leds

    def _update_leds(self, group, value):
        self._states[group].value = value


class DimmableLightingControl_KeypadLinc_6(DimmableLightingControl_KeypadLinc):
    """KeypadLinc 6 button dimmer."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the DimmableLightingControl_KeypadLinc_6 class."""
        buttons = {
            1: DIMMABLE_LIGHT_MAIN,
            3: ON_OFF_SWITCH_A,
            4: ON_OFF_SWITCH_B,
            5: ON_OFF_SWITCH_C,
            6: ON_OFF_SWITCH_D,
        }
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )


class DimmableLightingControl_KeypadLinc_8(DimmableLightingControl_KeypadLinc):
    """KeypadLinc 8 button dimmer."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the DimmableLightingControl_KeypadLinc_6 class."""
        buttons = {
            1: DIMMABLE_LIGHT_MAIN,
            2: ON_OFF_SWITCH_B,
            3: ON_OFF_SWITCH_C,
            4: ON_OFF_SWITCH_D,
            5: ON_OFF_SWITCH_E,
            6: ON_OFF_SWITCH_F,
            7: ON_OFF_SWITCH_G,
            8: ON_OFF_SWITCH_H,
        }
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )



"""Receive an All-Link record for an Insteon Modem."""

import logging

from .inbound_base import InboundHandlerBase
from ..address import Address
from ..topics import ALL_LINK_RECORD_RESPONSE
from . import inbound_handler
from ..protocol.messages.all_link_record_flags import AllLinkRecordFlags

_LOGGER = logging.getLogger(__name__)


class AllLinkRecordResponseHandler(InboundHandlerBase):
    """Receive an All-Link record for an Insteon Modem."""

    def __init__(self):
        """Init the AllLinkRecordResponse class."""
        super().__init__(ALL_LINK_RECORD_RESPONSE)

    @inbound_handler
    def receive_record(
        self,
        flags: AllLinkRecordFlags,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
    ):
        """Recieve an all link record."""
        from ..constants import AllLinkMode

        controller = flags.mode == AllLinkMode.CONTROLLER
        self._call_subscribers(
            in_use=flags.is_in_use,
            high_water_mark=flags.is_hwm,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            bit5=flags.is_bit_5_set,
            bit4=flags.is_bit_4_set,
        )



"""Manage outbound ON command to a device."""
from ... import pub
from .. import status_handler
from .direct_command import DirectCommandHandlerBase
from ...topics import STATUS_REQUEST


class StatusRequestCommand(DirectCommandHandlerBase):
    """Manage an outbound Status command to a device.

    TODO Confirm that the status command is best with a single status handler
    rather than a handler per status command (ie. one for state 1 and one for
    state 2)
    """

    def __init__(self, address, status_type: int = 0):
        """Init the OnLevelCommand class."""
        super().__init__(topic=STATUS_REQUEST, address=address)
        self._status_type = status_type

    # pylint: disable=arguments-differ, useless-super-delegation
    def send(self):
        """Send the ON command."""
        super().send(status_type=self._status_type)

    # pylint: disable=arguments-differ, useless-super-delegation
    async def async_send(self):
        """Send the ON command async."""
        return await super().async_send(status_type=self._status_type)

    @status_handler
    def handle_direct_ack(self, topic=pub.AUTO_TOPIC, **kwargs):
        """Handle the ON response direct ACK."""
        cmd1 = kwargs.get("cmd1")
        cmd2 = kwargs.get("cmd2")
        if cmd2 is not None:
            self._call_subscribers(db_version=cmd1, status=cmd2)



"""Manage inbound ON command from device."""
from .. import broadcast_handler
from .broadcast_command import BroadcastCommandHandlerBase
from ...topics import OFF_FAST
from ...address import Address


class OffFastInbound(BroadcastCommandHandlerBase):
    """Off Level command inbound."""

    def __init__(self, address, group):
        """Init the OffFastInbound class."""
        self._address = Address(address)
        self._group = group
        super().__init__(topic=OFF_FAST, address=self._address, group=self._group)

    @broadcast_handler
    def handle_command(self, cmd1, cmd2, target, user_data):
        """Handle the OFF command from a device."""
        self._call_subscribers(on_level=0)



"""Manage links beteween devices."""
import logging
from ..device_types.device_base import Device
from .. import devices
from ..constants import AllLinkMode, ResponseStatus
from ..handlers.start_all_linking import StartAllLinkingCommandHandler
from ..utils import multiple_status

TIMEOUT = 3
_LOGGER = logging.getLogger(__name__)


async def async_enter_linking_mode(is_controller: bool, group: int):
    """Put the Insteon Modem into linking mode."""
    link_cmd = StartAllLinkingCommandHandler()
    mode = AllLinkMode.CONTROLLER if is_controller else AllLinkMode.RESPONDER
    response = await link_cmd.async_send(mode=mode, group=group)
    _LOGGER.debug("Enter linking mode response: %s", str(response))
    return response


async def async_enter_unlinking_mode(group: int):
    """Put the Insteon Modem into unlinking mode."""
    link_cmd = StartAllLinkingCommandHandler()
    mode = AllLinkMode.DELETE
    response = await link_cmd.async_send(mode=mode, group=group)
    _LOGGER.info("Enter linking mode response: %s", str(response))
    return response


async def async_link_devices(
    controller: Device, responder: Device, group: int = 0, data1=255, data2=28, data3=1
):
    """Link two devices."""
    if _add_link_to_device(
        device=controller,
        target=responder.address,
        group=group,
        is_controller=True,
        data1=responder.cat,
        data2=responder.subcat,
        data3=responder.firmware,
    ) and _add_link_to_device(
        device=responder,
        target=controller.address,
        group=group,
        is_controller=False,
        data1=data1,
        data2=data2,
        data3=data3,
    ):
        _, failed_1 = await controller.aldb.async_write()
        _, failed_2 = await responder.aldb.async_write()
        if failed_1 or failed_2:
            return ResponseStatus.FAILURE
        return ResponseStatus.SUCCESS


async def async_unlink_devices(controller: Device, responder: Device, group: int = 0):
    """Unlink two devices."""
    if _remove_link_from_device(
        device=controller, is_controller=True, group=group, target=responder.address
    ) and _remove_link_from_device(
        device=responder, is_controller=False, group=group, target=controller.address
    ):
        _, failed_1 = await controller.aldb.async_write()
        _, failed_2 = await responder.aldb.async_write()
        if failed_1 or failed_2:
            return ResponseStatus.FAILURE
        return ResponseStatus.SUCCESS
    return ResponseStatus.FAILURE


async def async_add_default_links(device: Device):
    """Set up the default links for a device."""
    for link in device.default_links:
        if link.is_controller:
            controller = device
            responder = devices.modem
            data1 = link.modem_data1
            data2 = link.modem_data2
            data3 = link.modem_data3
        else:
            controller = devices.modem
            responder = device
            data1 = link.dev_data1
            data2 = link.dev_data2
            data3 = link.dev_data3
        await async_link_devices(controller, responder, link.group, data1, data2, data3)


def _remove_link_from_device(device, is_controller, group, target):
    found_rec = None
    for addr in device.aldb:
        rec = device.aldb[addr]
        if (
            rec.target == target
            and rec.is_controller == is_controller
            and rec.group == group
        ):
            found_rec = rec
    if found_rec:
        device.aldb.remove(found_rec.mem_addr)
    return True


async def async_create_default_links(device: Device):
    """Establish default links between the modem and device."""

    if not device.aldb.is_loaded:
        await device.aldb.async_load()

    if not device.aldb.is_loaded:
        return False

    results = []
    for link_info in device.default_links:
        is_controller = link_info.is_controller
        group = link_info.group
        data1 = link_info.data1
        data2 = link_info.data2
        data3 = link_info.data3
        if is_controller:
            controller = device
            responder = devices.modem
        else:
            controller = devices.modem
            responder = device
        result = await async_link_devices(
            controller, responder, group, data1, data2, data3
        )
        results.append(result)
    return multiple_status(*results)


def _add_link_to_device(device, is_controller, group, target, data1, data2, data3):
    """Add a link to a device."""
    try:
        device.aldb.add(
            group=group,
            target=target,
            controller=is_controller,
            data1=data1,
            data2=data2,
            data3=data3,
        )
    except NotImplementedError:
        return False
    return True



"""Handle sending a read request for ALDB records."""
import logging

from ...address import Address
from ...topics import SET_OPERATING_FLAGS
from .direct_command import DirectCommandHandlerBase

_LOGGER = logging.getLogger(__name__)


class SetOperatingFlagsCommand(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the ReadALDBCommandHandler."""
        super().__init__(topic=SET_OPERATING_FLAGS, address=address)

    # pylint: disable=arguments-differ
    def send(self, cmd: int):
        """Send Get Operating Flags message."""
        super().send(cmd=cmd)

    # pylint: disable=arguments-differ
    async def async_send(self, cmd: int):
        """Send Get Operating Flags message asyncronously."""
        return await super().async_send(cmd=cmd)



"""Manage outbound ON command to a device."""

from .. import direct_ack_handler
from .direct_command import DirectCommandHandlerBase
from ...topics import ON


class OnLevelCommand(DirectCommandHandlerBase):
    """Manage an outbound ON command to a device."""

    def __init__(self, address, group):
        """Init the OnLevelCommand class."""
        super().__init__(topic=ON, address=address, group=group)

    @property
    def group(self):
        """Command group."""
        return self._group

    # pylint: disable=arguments-differ
    def send(self, on_level=0xFF):
        """Send the ON command."""
        super().send(on_level=on_level, group=self._group)

    # pylint: disable=arguments-differ
    async def async_send(self, on_level=0xFF):
        """Send the ON command async."""
        return await super().async_send(on_level=on_level, group=self._group)

    @direct_ack_handler
    def handle_direct_ack(self, cmd1, cmd2, target, user_data):
        """Handle the ON response direct ACK."""
        if self._response_lock.locked():
            self._call_subscribers(on_level=cmd2 if cmd2 else 0xFF)



"""Switched Lighting Control devices (CATEGORY 0x02)."""
from ..events import OFF_EVENT, OFF_FAST_EVENT, ON_EVENT, ON_FAST_EVENT
from ..handlers.to_device.set_leds import SetLedsCommandHandler

# from ..handlers.to_device.trigger_scene_on import TriggerSceneOnCommandHandler
# from ..handlers.to_device.trigger_scene_off import TriggerSceneOffCommandHandler
from ..handlers.to_device.status_request import StatusRequestCommand
from ..states import (
    ON_OFF_OUTLET_BOTTOM,
    ON_OFF_OUTLET_TOP,
    ON_OFF_SWITCH,
    ON_OFF_SWITCH_A,
    ON_OFF_SWITCH_B,
    ON_OFF_SWITCH_C,
    ON_OFF_SWITCH_D,
    ON_OFF_SWITCH_E,
    ON_OFF_SWITCH_F,
    ON_OFF_SWITCH_G,
    ON_OFF_SWITCH_H,
    ON_OFF_SWITCH_MAIN,
)
from ..states.on_off import OnOff
from .commands import SET_LEDS_COMMAND, STATUS_COMMAND
from .on_off_responder_base import OnOffResponderBase


class SwitchedLightingControl(OnOffResponderBase):
    """Switched Lighting Control device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
        state_name=ON_OFF_SWITCH,
        on_event_name=ON_EVENT,
        off_event_name=OFF_EVENT,
        on_fast_event_name=ON_FAST_EVENT,
        off_fast_event_name=OFF_FAST_EVENT,
    ):
        """Init the OnOffResponderBase class."""
        buttons = {1: ON_OFF_SWITCH} if buttons is None else buttons
        super().__init__(
            address,
            cat,
            subcat,
            firmware,
            description,
            model,
            buttons,
            on_event_name,
            off_event_name,
            on_fast_event_name,
            off_fast_event_name,
        )


class SwitchedLightingControl_ApplianceLinc(SwitchedLightingControl):
    """ApplianceLinc based dimmable lights."""

    def _register_operating_flags(self):
        from ..operating_flag import PROGRAM_LOCK_ON, LED_BLINK_ON_TX_ON, LED_ON

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9)


class SwitchedLightingControl_SwitchLinc(SwitchedLightingControl):
    """SwichLinc based dimmable lights."""

    def _register_operating_flags(self):
        from ..operating_flag import (
            PROGRAM_LOCK_ON,
            LED_BLINK_ON_TX_ON,
            RESUME_DIM_ON,
            LED_ON,
            KEY_BEEP_ON,
            LED_BLINK_ON_ERROR_ON,
        )
        from ..extended_property import (
            LED_DIMMING,
            ON_LEVEL,
            X10_HOUSE,
            X10_UNIT,
            RAMP_RATE,
        )

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)
        self._add_operating_flag(LED_BLINK_ON_ERROR_ON, 5, 2, 0x14, 0x15)

        self._add_property(LED_DIMMING, 3, 3)
        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)
        self._add_property(RAMP_RATE, 7, 5)
        self._add_property(ON_LEVEL, 8, 6)


class SwitchedLightingControl_InLineLinc(SwitchedLightingControl_SwitchLinc):
    """InLineLinc based dimmable lights."""


class SwitchedLightingControl_OutletLinc(SwitchedLightingControl):
    """OutletLinc based dimmable lights."""

    def _register_operating_flags(self):
        from ..operating_flag import PROGRAM_LOCK_ON, LED_BLINK_ON_TX_ON, LED_ON
        from ..extended_property import X10_HOUSE, X10_UNIT

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9)

        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)


class SwitchedLightingControl_Micro(SwitchedLightingControl):
    """Micro switch based dimmable lights."""

    def _register_operating_flags(self):
        from ..operating_flag import (
            PROGRAM_LOCK_ON,
            LED_BLINK_ON_TX_ON,
            LED_ON,
            KEY_BEEP_ON,
            DUAL_LINE_ON,
            MOMENTARY_LINE_ON,
            THREE_WAY_ON,
            REVERSED_ON,
            LED_BLINK_ON_ERROR_ON,
        )

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)

        self._add_operating_flag(LED_BLINK_ON_ERROR_ON, 2, 2, 0x15, 0x14)

        self._add_operating_flag(DUAL_LINE_ON, 3, 0, 0x1E, 0x1F)
        self._add_operating_flag(MOMENTARY_LINE_ON, 3, 1, 0x20, 0x21)
        self._add_operating_flag(THREE_WAY_ON, 3, 2, 0x23, 0x22)
        self._add_operating_flag(REVERSED_ON, 3, 3, 0x25, 0x24)


class SwitchedLightingControl_DinRail(SwitchedLightingControl):
    """DINRail based dimmable lights."""

    def _register_operating_flags(self):
        from ..operating_flag import (
            PROGRAM_LOCK_ON,
            LED_BLINK_ON_TX_ON,
            LED_ON,
            KEY_BEEP_ON,
        )
        from ..extended_property import (
            LED_DIMMING,
            ON_LEVEL,
            X10_HOUSE,
            X10_UNIT,
            RAMP_RATE,
        )

        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)

        self._add_property(LED_DIMMING, 3, 3)
        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)
        self._add_property(RAMP_RATE, 7, 5)
        self._add_property(ON_LEVEL, 8, 6)


class SwitchedLightingControl_KeypadLinc(SwitchedLightingControl):
    """KeypadLinc base class."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the SwitchedLightingControl_KeypadLinc class."""
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )

    async def async_on(self, group: int = 0):
        """Turn on the button LED."""
        if group in [0, 1]:
            return await super().async_on(group=group)
        kwargs = self._change_led_status(led=group, on=True)
        return await self._handlers[SET_LEDS_COMMAND].async_send(**kwargs)

    async def async_off(self, group: int = 0):
        """Turn off the button LED."""
        if group in [0, 1]:
            return await super().async_off(group=group)
        kwargs = self._change_led_status(led=group, on=False)
        return await self._handlers[SET_LEDS_COMMAND].async_send(**kwargs)

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[SET_LEDS_COMMAND] = SetLedsCommandHandler(address=self.address)

    def _register_states(self):
        super()._register_states()
        for button in self._buttons:
            name = self._buttons[button]
            self._states[button] = OnOff(name=name, address=self._address, group=button)

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._handlers[SET_LEDS_COMMAND].subscribe(self._update_leds)

    def _change_led_status(self, led, on):
        leds = {}
        for curr_led in range(1, 9):
            var = "group{}".format(curr_led)
            leds[var] = on if curr_led == led else bool(self._states.get(curr_led))
        return leds

    def _update_leds(self, group, value):
        self._states[group].value = value


class SwitchedLightingControl_KeypadLinc_6(SwitchedLightingControl_KeypadLinc):
    """KeypadLinc 6 button switch."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the SwitchedLightingControl_KeypadLinc_6 class."""
        buttons = {
            1: ON_OFF_SWITCH_MAIN,
            3: ON_OFF_SWITCH_A,
            4: ON_OFF_SWITCH_B,
            5: ON_OFF_SWITCH_C,
            6: ON_OFF_SWITCH_D,
        }
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )


class SwitchedLightingControl_KeypadLinc_8(SwitchedLightingControl_KeypadLinc):
    """KeypadLinc 8 button switch."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the SwitchedLightingControl_KeypadLinc_8 class."""
        buttons = {
            1: ON_OFF_SWITCH_MAIN,
            2: ON_OFF_SWITCH_B,
            3: ON_OFF_SWITCH_C,
            4: ON_OFF_SWITCH_D,
            5: ON_OFF_SWITCH_E,
            6: ON_OFF_SWITCH_F,
            7: ON_OFF_SWITCH_G,
            8: ON_OFF_SWITCH_H,
        }
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            button=buttons,
        )


class SwitchedLightingControl_OnOffOutlet(SwitchedLightingControl_ApplianceLinc):
    """On/Off outlet model 2663-222 Switched Lighting Control.

    Device Class 0x02 subcat 0x39
    """

    TOP_GROUP = 1
    BOTTOM_GROUP = 2

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the SwitchedLightingControl_KeypadLinc class."""
        buttons = {1: ON_OFF_OUTLET_TOP, 2: ON_OFF_OUTLET_BOTTOM}
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )

    def status(self, group=None):
        """Request the status of the device."""
        if group is None:
            self._handlers[STATUS_COMMAND].send()
        if group in [1, 2]:
            self._handlers[group][STATUS_COMMAND].send()

    async def async_status(self, group=None):
        """Request the status of the device."""
        if group is None:
            return await self._handlers[STATUS_COMMAND].async_send()
        if group in [1, 2]:
            return await self._handlers[group][STATUS_COMMAND].async_send()

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[STATUS_COMMAND] = StatusRequestCommand(self._address, 1)

        if self._handlers.get(self.TOP_GROUP) is None:
            self._handlers[self.TOP_GROUP] = {}
        self._handlers[self.TOP_GROUP][STATUS_COMMAND] = StatusRequestCommand(
            self._address, 0
        )

        if self._handlers.get(self.BOTTOM_GROUP) is None:
            self._handlers[self.BOTTOM_GROUP] = {}
        self._handlers[self.BOTTOM_GROUP][STATUS_COMMAND] = StatusRequestCommand(
            self._address, 0
        )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()

        self._handlers[self.TOP_GROUP][STATUS_COMMAND].subscribe(
            self._handle_top_status
        )
        self._handlers[self.TOP_GROUP][STATUS_COMMAND].subscribe(
            self._handle_bottom_status
        )

    def _handle_status(self, db_version, status):
        """Set the status of the top and bottom outlets state."""
        self._states[self.TOP_GROUP].value = status & 0x02
        self._states[self.BOTTOM_GROUP].value = status & 0x01

    def _handle_top_status(self, db_version, status):
        """Set the status of the top outlet."""
        self._states[self.TOP_GROUP].value = status

    def _handle_bottom_status(self, db_version, status):
        """Set the status of the bottom outlet."""
        self._states[self.BOTTOM_GROUP].value = status



"""Wet / Dry state."""
from . import StateBase
from ..address import Address


class WetDry(StateBase):
    """Wet / Dry state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: int = None
    ):
        """Init the WetDry class."""
        super().__init__(name, address, group, default, value_type=bool)

    # pylint: disable=arguments-differ
    def set_value(self, wet):
        """Set the value of the state from the handlers."""
        self.value = bool(wet)
